// Control.cpp -- IOCTL handlers for AIOWDM driver

#include "stddcls.h"
#include "driver.h"
#include "ioctls.h"

void ForgetIRP(PIRP IRP);
NTSTATUS DispatchDMAPiongControl(PDEVICE_EXTENSION pDX, PIRP pIRP, PUCHAR pCmd, ULONG InSize, ULONG OutSize, ULONG & DataOut);

VOID IRQGenCancelIrp(IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp);

///////////////////////////////////////////////////////////////////////////////

#pragma PAGEDCODE

NTSTATUS DispatchControl(PDEVICE_OBJECT fdo, PIRP Irp)
{							// DispatchControl
	PAGED_CODE();
	PDEVICE_EXTENSION pdx = (PDEVICE_EXTENSION) fdo->DeviceExtension;

	NTSTATUS status = STATUS_SUCCESS;
	if ( pdx->bRealDevice )
	{
		status = IoAcquireRemoveLock(&pdx->RemoveLock, Irp);
		if ( ! NT_SUCCESS(status) )
			return CompleteRequest(Irp, status, 0)
		;
	}
	ULONG info = 0;

	PIO_STACK_LOCATION stack = IoGetCurrentIrpStackLocation(Irp);
	ULONG cbin = stack->Parameters.DeviceIoControl.InputBufferLength;
	ULONG cbout = stack->Parameters.DeviceIoControl.OutputBufferLength;
	ULONG code = stack->Parameters.DeviceIoControl.IoControlCode;

	switch ( code )
	{						// by IOCTL code

		case IOCTL_QUERY_INFO:				// code == 0x800
		{						// IOCTL_QUERY_INFO

			if ( ! pdx->bRealDevice )
			{
				status = STATUS_OBJECT_TYPE_MISMATCH;
				break;
			}

			if (cbout < sizeof(TCARDINFO))
			{
				status = STATUS_INVALID_PARAMETER;
				break;
			}

			PCARDINFO CIBuf = PCARDINFO(Irp->AssociatedIrp.SystemBuffer);
			CIBuf->DeviceID = pdx->DeviceID;
			CIBuf->BaseAddress = ULONG(pdx->MainBar.Base);
			info = sizeof(TCARDINFO);

			status = STATUS_SUCCESS;

		break;
		}						// IOCTL_QUERY_INFO

		case IOCTL_QUERY_BAR_BASE:				// code == 0x801
		{						// IOCTL_QUERY_BAR_BASE

			if ( ! pdx->bRealDevice )
			{
				status = STATUS_OBJECT_TYPE_MISMATCH;
				break;
			}

			if (cbout < sizeof(ULONG))
			{
				status = STATUS_INVALID_PARAMETER;
				break;
			}

			PULONG pBuf = PULONG(Irp->AssociatedIrp.SystemBuffer);
			switch ( *pBuf )
			{
				case 1:
					*pBuf = ULONG(pdx->StatusBar.Base);
				break;
				case 2:
					*pBuf = ULONG(pdx->MainBar.Base);
				break;
				case 3:
					*pBuf = ULONG(pdx->AuxBar.Base);
				break;
				default:
					*pBuf = 0;
				//break;
			}
			info = sizeof(*pBuf);
			status = STATUS_SUCCESS;

		break;
		}						// IOCTL_QUERY_BAR_BASE

		case IOCTL_PORT_IN:				// code == 0x820
		{						// IOCTL_PORT_IN

			if ( (cbin < sizeof(TPORTOP)) || cbout < sizeof(TPORTOP) )
			{
				status = STATUS_INVALID_PARAMETER;
				break;
			}
			PPORTOP POBuf = PPORTOP(Irp->AssociatedIrp.SystemBuffer);
			TBarData TarBar = pdx->MainBar;
			switch ( POBuf->Size )
			{
				case 1:
					POBuf->Data8 = InBar8(TarBar, POBuf->Address);
				break;
				case 2:
					POBuf->Data16 = InBar16(TarBar, POBuf->Address);
				break;
				case 4:
					POBuf->Data32 = InBar32(TarBar, POBuf->Address);
				break;
			}
			info = sizeof(TPORTOP);

			status = STATUS_SUCCESS;

		break;
		}						// IOCTL_PORT_IN

		case IOCTL_PORT_INSTR:				// code == 0x821
		{						// IOCTL_PORT_INSTR

			if ( (cbin < sizeof(TPORTOP)) || cbout < sizeof(TPORTOP) )
			{
				status = STATUS_INVALID_PARAMETER;
				break;
			}
			PPORTOP POBuf = PPORTOP(Irp->AssociatedIrp.SystemBuffer);
			TBarData TarBar = pdx->MainBar;
			switch ( POBuf->Size )
			{
				case 1:
					InBarStr8(TarBar, POBuf->Address, (PUCHAR)Irp->AssociatedIrp.SystemBuffer, cbout);
				break;
				case 2:
					InBarStr16(TarBar, POBuf->Address, (PUSHORT)Irp->AssociatedIrp.SystemBuffer, cbout / 2);
				break;
				case 4:
					InBarStr32(TarBar, POBuf->Address, (PULONG)Irp->AssociatedIrp.SystemBuffer, cbout / 4);
				break;
			}
			info = cbout;

			status = STATUS_SUCCESS;

		break;
		}						// IOCTL_PORT_INSTR

		case IOCTL_PORT_OUT:				// code == 0x822
		{						// IOCTL_PORT_OUT

			if (cbout < sizeof(TPORTOP))
			{
				status = STATUS_INVALID_PARAMETER;
				break;
			}
			PPORTOP POBuf = PPORTOP(Irp->AssociatedIrp.SystemBuffer);
			TBarData TarBar = pdx->MainBar;
			switch ( POBuf->Size )
			{
				case 1:
					OutBar8(TarBar, POBuf->Address, POBuf->Data8);
				break;
				case 2:
					OutBar16(TarBar, POBuf->Address, POBuf->Data16);
				break;
				case 4:
					OutBar32(TarBar, POBuf->Address, POBuf->Data32);
				break;
			}
			info = 0;

			status = STATUS_SUCCESS;

		break;
		}						// IOCTL_PORT_OUT

		case IOCTL_PORTMB_OUT:
		{						// IOCTL_PORTMB_OUT

			if ( cbout < sizeof(TPORTOP)-4+1 )
			{
				status = STATUS_INVALID_PARAMETER;
				break;
			}
			PPORTOP POBuf = PPORTOP(Irp->AssociatedIrp.SystemBuffer);
			if ( cbout < sizeof(TPORTOP)-4+POBuf->Size )
			{
				status = STATUS_INVALID_PARAMETER;
				break;
			}
			TBarData TarBar = pdx->MainBar;
			OutBar8M(TarBar, POBuf->Address, POBuf->Size, &POBuf->Data8);
			info = 0;

			status = STATUS_SUCCESS;

		break;
		}						// IOCTL_PORTMB_OUT

		case IOCTL_IRQ_WAIT_BEGIN:				// code == 0x810
		{						// IOCTL_IRQ_WAIT_BEGIN

			if ( ! pdx->bRealDevice )
			{
				status = STATUS_OBJECT_TYPE_MISMATCH;
				break;
			}

			if (pdx->InterruptRequested)
			{
				status = STATUS_INVALID_DEVICE_REQUEST;
			}
			else if (pdx->InterruptOccurred)
			{
				pdx->InterruptOccurred = FALSE;
				if ( cbout )
				{
					InCOSData(pdx, Irp, cbout);
					info = cbout;
				}
				status = STATUS_SUCCESS;
			}
			else
			{
				KIRQL OldIRQL;

				pdx->COSSize = cbout;
				pdx->PendingIrp = Irp;
				pdx->HasPendingIrp = TRUE;
				IoMarkIrpPending(Irp);
				IoAcquireCancelSpinLock(&OldIRQL);
				IoSetCancelRoutine(Irp, IRQGenCancelIrp);
				IoReleaseCancelSpinLock(OldIRQL);
				pdx->InterruptRequested = TRUE;
				status = STATUS_PENDING;   // do not complete Irp of course
			}

		break;
		}						// IOCTL_IRQ_WAIT_BEGIN

		case IOCTL_IRQ_WAIT_ABORT:				// code == 0x812
		{

			if ( ! pdx->bRealDevice )
			{
				status = STATUS_OBJECT_TYPE_MISMATCH;
				break;
			}

			if (!pdx->InterruptRequested)
				status = STATUS_INVALID_DEVICE_REQUEST;
			else
			{
				KIRQL OldIRQL;

				pdx->InterruptOccurred = FALSE;
				pdx->InterruptRequested = FALSE;
				pdx->HasPendingIrp = FALSE;
				pdx->PendingIrp->IoStatus.Status = STATUS_CANCELLED;
				pdx->PendingIrp->IoStatus.Information = 0;
				IoAcquireCancelSpinLock(&OldIRQL);
				IoSetCancelRoutine(pdx->PendingIrp, NULL);
				IoReleaseCancelSpinLock(OldIRQL);
				IoCompleteRequest(pdx->PendingIrp, IO_SERIAL_INCREMENT);
				status = STATUS_SUCCESS;
			}

		break;
		}

		case IOCTL_SET_COS_SCHEME:
		{
			if ( (cbin < sizeof(ULONG)) || (cbout < sizeof(ULONG)) )
			{
				status = STATUS_INVALID_BUFFER_SIZE;
				break;
			}

			ULONG COSScheme = *PULONG(Irp->AssociatedIrp.SystemBuffer);
			if ( (COSScheme & 0xF0808080) != 0x10000000 )
			{
				status = STATUS_NOT_IMPLEMENTED;
				break;
			}

			//In the future when new versions have been added, case on version here.

			ULONG PPIStartOff = COSScheme & 0x7F;
			ULONG PPISize = (COSScheme >> 8) & 0x7F;
			ULONG PPIStride = (COSScheme >> 16) & 0x7F;

			if ( PPISize == 0 )
			{
				status = STATUS_INVALID_PARAMETER;
				break;
			}

			if ( ! pdx->bRealDevice )
			{
				status = STATUS_OBJECT_TYPE_MISMATCH;
				break;
			}

			if (pdx->InterruptRequested)
			{
				status = STATUS_INVALID_DEVICE_REQUEST;
				break;
			}

			pdx->COSScheme.StartOff = UCHAR(PPIStartOff);
			pdx->COSScheme.Size = UCHAR(PPISize);
			pdx->COSScheme.Stride = UCHAR(PPIStride);
 
			*PULONG(Irp->AssociatedIrp.SystemBuffer) = PPISize;
			info = sizeof(PPISize);
		break;
		}

		case IOCTL_FORGET_IRP:				//code == 0x880
		{

			if ( ! pdx->bRealDevice )
				status = STATUS_OBJECT_TYPE_MISMATCH;
			else if ( ! pdx->InterruptRequested )
				status = STATUS_NO_SUCH_FILE;
			else
			{
				pdx->InterruptRequested = FALSE;
				pdx->HasPendingIrp = FALSE;
				ForgetIRP(pdx->PendingIrp);
				pdx->PendingIrp = NULL;
				status = STATUS_SUCCESS;
			}

		break;
		}

		case IOCTL_IRQ_POLL:				// code == 0x814
		{

			if ( ! pdx->bRealDevice )
			{
				status = STATUS_OBJECT_TYPE_MISMATCH;
				break;
			}

			//if ( InterlockedExchange(PLONG(&(pdx->InterruptOccurred)), 0) )
			if ( pdx->InterruptOccurred )
			{
				pdx->InterruptOccurred = 0;
				status = STATUS_SUCCESS;
			}
			else
				status = STATUS_NO_MORE_FILES
			;

		break;
		}

		case IOCTL_DMA_PIONG_CONTROL:				// code == 0x840
		{

			if ( ! pdx->bRealDevice )
			{
				status = STATUS_OBJECT_TYPE_MISMATCH;
				break;
			}

			if ( cbin < 1 )
			{
				status = STATUS_BUFFER_TOO_SMALL;
				break;
			}

			if ( (pdx->DMALink[0].pBuffer == NULL) || (pdx->DMALink[1].pBuffer == NULL) )
			{
				status = STATUS_ORDINAL_NOT_FOUND;
				break;
			}

			status = DispatchDMAPiongControl(pdx, Irp, PUCHAR(Irp->AssociatedIrp.SystemBuffer), cbin, cbout, info);
		break;
		}

		case IOCTL_DMA_PIONG_READ:				// code == 0x840
		{

			if ( ! pdx->bRealDevice )
			{
				status = STATUS_OBJECT_TYPE_MISMATCH;
				break;
			}

			if ( ! pdx->bDMAGoing )
			{
				status = STATUS_DEVICE_NOT_READY;
				break;
			}

			if ( cbout < pdx->DMASizeUsed )
			{
				status = STATUS_BUFFER_TOO_SMALL;
				break;
			}

			KIRQL OldIRQL;

			TPDMAUserRead pNewRead = TPDMAUserRead(ExAllocatePool(NonPagedPool, sizeof(*pNewRead)));

			MmProbeAndLockPages(Irp->MdlAddress, KernelMode, IoWriteAccess);
			pNewRead->pBuffer = MmGetSystemAddressForMdlSafe(Irp->MdlAddress, NormalPagePriority);
			pNewRead->BufferSize = pdx->DMASizeUsed; //Maybe someday this won't be 1:1, and we can use cbout here instead.
			pNewRead->pPendingIRP = Irp;
			pNewRead->pNext = NULL;

			IoMarkIrpPending(Irp);
			IoAcquireCancelSpinLock(&OldIRQL);
			IoSetCancelRoutine(Irp, IRQGenCancelIrp); //`@
			IoReleaseCancelSpinLock(OldIRQL);
			status = STATUS_PENDING;

			KeRaiseIrql(pdx->MyDIRQL, &OldIRQL);
			if ( pdx->pLastRead )
				pdx->pLastRead->pNext = pNewRead;
			else
				pdx->pFirstRead = pNewRead
			;
			pdx->pLastRead = pNewRead;
			KeLowerIrql(OldIRQL);
		break;
		}

		case IOCTL_GET_BENCH:				// code == 0x830
		{

			if ( ! pdx->bRealDevice )
			{
				status = STATUS_OBJECT_TYPE_MISMATCH;
				break;
			}

			if (cbout < 4)
			{
				status = STATUS_INVALID_PARAMETER;
				break;
			}

			info = 4;
			*PULONG(Irp->AssociatedIrp.SystemBuffer) = pdx->Bench;
			pdx->Bench = 0;
			status = STATUS_SUCCESS;

		break;
		}

/*
		case IOCTL_`:				// code == 0x`
		{

			if ( ! pdx->bRealDevice )
			{
				status = STATUS_OBJECT_TYPE_MISMATCH;
				break;
			}

			`

		break;
		}
*/

		default:
		{
			DbgPrint("Unknown IOCTL = %08LX\n", code);

			LOCAL_UNICODE_STRING(DebugData, 256);
			UStrAddStr(DebugData, L"Unknown IOCTL = ");
			UStrAdd32(DebugData, code);
			WriteRegStr(UDebugKey, iLogEvent++, DebugData);

			status = STATUS_INVALID_DEVICE_REQUEST;
		break;
		}

	}						// by IOCTL code

	if ( pdx->bRealDevice ) IoReleaseRemoveLock(&pdx->RemoveLock, Irp);
	if (status == STATUS_PENDING)
		return status;
	else
		return CompleteRequest(Irp, status, info)
	;
}							// DispatchControl

///////////////////////////////////////////////////////////////////////////////

#pragma PAGEDCODE

NTSTATUS DispatchDMAPiongControl(PDEVICE_EXTENSION pDX, PIRP pIRP, PUCHAR pCmd, ULONG InSize, ULONG OutSize, ULONG & DataOut)
{
	PAGED_CODE();

	DataOut = 0;

	switch ( pCmd[0] )
	{
		case 0x02:
		{
			ULONG BufferSize = DMA_BUFFER_SIZE;
			if ( OutSize < sizeof(BufferSize) ) return STATUS_BUFFER_TOO_SMALL;

			*PULONG(pCmd) = BufferSize;
			DataOut = sizeof(BufferSize);
		break;
		}
		case 0x10:
		{
			BOOLEAN bWasStarted;
			TPDMAUserRead pCancelRead = NULL;
			TPDMAUserRead pCancelFreeRead = NULL;

			KIRQL OldIRQL;
			KeRaiseIrql(pDX->MyDIRQL, &OldIRQL);
			bWasStarted = pDX->bDMAGoing;
			if ( bWasStarted )
			{
				pDX->bDMAGoing = FALSE;
				pCancelRead = pDX->pFirstRead;
				pDX->pFirstRead = NULL;
				pDX->pLastRead = NULL;
				pCancelFreeRead = pDX->pFirstFreeRead;
				pDX->pFirstFreeRead = NULL;
				pDX->pLastFreeRead = NULL;
			}
			KeLowerIrql(OldIRQL);

			DMAUserReadsFreeAndComplete(pCancelFreeRead, STATUS_TRANSLATION_COMPLETE); //I wish STATUS_SUCCESS worked
			DMAUserReadsFreeAndComplete(pCancelRead, STATUS_CANCELLED);

			if ( ! bWasStarted ) return STATUS_CONNECTION_INVALID;
		break;
		}
		case 0x11:
		{
			NTSTATUS Status = STATUS_SUCCESS;

			KIRQL OldIRQL;
			KeRaiseIrql(pDX->MyDIRQL, &OldIRQL);
			if ( pDX->bDMAGoing )
			{
				Status = STATUS_CONNECTION_ACTIVE;
			}
			else
			{
				pDX->DMASizeUsed = InBar32(pDX->MainBar, 0x20);
				if ( ! pDX->DMASizeUsed )
				{
					Status = STATUS_BUFFER_TOO_SMALL;
				}
				else
				{
					pDX->iDMALink = 0;
					pDX->bDMAWrote = FALSE;
					pDX->bDMAGoing = TRUE;
				}
			}
			KeLowerIrql(OldIRQL);

			return Status;
		break;
		}
/*
		case 0x`:
		{
			if ( InSize < 2 ) return STATUS_BUFFER_TOO_SMALL;

			ULONG iDMALink = PUCHAR(Irp->AssociatedIrp.SystemBuffer)[1];
			if ( iDMALink >= pdx->DMALinkCt ) return STATUS_TOO_MANY_LINKS;

			TPDMALink pLink = &(pdx->DMALink[iDMALink]);
			if ( pLink->pBuffer == NULL ) return STATUS_ORDINAL_NOT_FOUND;

			`
		break;
		}
*/
		default: return STATUS_ORDINAL_NOT_FOUND;
	}

	return STATUS_SUCCESS;
}

/*********************************************************************************
*      CANCEL IRP
**********************************************************************************/
VOID
IRQGenCancelIrp(
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP Irp
    )
{
	PDEVICE_EXTENSION extension = (PDEVICE_EXTENSION)DeviceObject->DeviceExtension;

	Irp->IoStatus.Information = 0;

	if (Irp == extension->PendingIrp)
	{
		extension->PendingIrp = NULL;
		extension->HasPendingIrp = FALSE;
		extension->InterruptRequested = FALSE;
		Irp->IoStatus.Information = 1;
	}
	IoReleaseCancelSpinLock(Irp->CancelIrql);

	Irp->IoStatus.Status = STATUS_CANCELLED;
	IoCompleteRequest(Irp, IO_NO_INCREMENT);
}